Program CRT_Bug_Fixer;
{$I-}
Uses DOS;

Const
  BugCode : Array [0..8] of Byte = ($F7,$D0,$F7,$D2,$B9,$37,$00,$F7,$F1);
  NewCode : Array [0..8] of Byte = ($F7,$D0,$F7,$D2,$B8,$37,$00,$90,$90);
  BufSize = 16384;

Var
  Buf : Array [0..BufSize-1+8] of Byte;
  Prog, Name, Backup : String;
  D : DirStr; N : NameStr; E : ExtStr;

Function UpStr (S : String) : String;
Var I : Byte;
Begin
  If S <> '' then
    For I := 1 to Length(S) do
      S[I] := UpCase(S[I]);
  UpStr := S
End;

Function Exists (Name : String) : Boolean;
Var
  F : File;
  Res : Boolean;
Begin
  FileMode := 0;
  Assign (F, Name);
  Reset (F, 1);
  Res := IOResult = 0;
  If Res then Close (F);
  FileMode := 2;
  Exists := Res
End;

Function CreateBackupCopy : Boolean;
Var
  F : File;
Begin
  Backup := Name;
  While Pos('.', Backup) <> 0 do Dec(Backup[0]);
  Assign (F, Name);
  Backup := Backup + '.BAK';
  Rename (F, Backup);
  CreateBackupCopy := IOResult = 0
End;

Function Scan (Read : Word) : Integer; Assembler;
Asm
  Mov   AX, DS
  Mov   ES, AX
  CLD
  Mov   BX, Read
  LEA   DX, Buf
@@cycle:
  LEA   SI, BugCode
  Mov   DI, DX
  Mov   CX, 9
  RepE  CMPSB
  JE    @@found
  Inc   DX
  Dec   BX
  JNZ   @@cycle
  Mov   AX, -1
  Jmp   @@end
@@found:
  Mov   AX, DX
  Sub   AX, Offset Buf
@@end:
End;

Function Fix : Boolean;
Var
  Res, Fixed : Boolean;
  FI, FO : File;
  Read, Written : Word;
  ScanRes : Integer;
  P1, P2,
  P : Pointer;
  Offs : LongInt;
Label
  LErr, LEnd;
Begin
  Res := False;
  Fixed := False;

  FileMode := 0;
  Assign (FI, Backup);
  Reset (FI, 1);
  If IOResult <> 0 then Goto LEnd;
  FileMode := 2;
  Assign (FO, Name);
  Rewrite (FO, 1);
  If IOResult <> 0 then Begin
    Close (FI);
    Goto LEnd
  End;

  FillChar (Buf, SizeOf(Buf), 0);
  P1 := @Buf; Inc (Word(P1), 8);
  P2 := @Buf; Inc (Word(P2), SizeOf(Buf)-8);

  Repeat
    BlockRead (FI, P1^, BufSize, Read);
    If IOResult <> 0 then Begin
LErr:
      Close (FI);
      Close (FO);
      Goto LEnd
    End;
    If Read = 0 then Break;
    ScanRes := Scan (Read);
    If ScanRes = -1 then Begin
      BlockWrite (FO, P1^, Read, Written);
      If (IOResult <> 0) or (Read <> Written) then Goto LErr
    End
    Else Begin
      P := @Buf; Inc (Word(P), ScanRes);
      Move (NewCode, P^, 9);
      Fixed := True;
      If ScanRes < 8 then Begin
        Offs := FilePos (FO);
        If IOResult <> 0 then Goto LErr;
        Seek (FO, Offs-(8-ScanRes));
        If IOResult <> 0 then Goto LErr;
        BlockWrite (FO, Buf, 8-ScanRes, Written);
        If (IOResult <> 0) or (8-ScanRes <> Written) then Goto LErr;
      End;
      BlockWrite (FO, P1^, Read, Written);
      If (IOResult <> 0) or (Read <> Written) then Goto LErr;
      Move (P2^, Buf, 8);
    End
  Until False;

  Close (FI);
  Close (FO);
  Res := IOResult = 0;
LEnd:
  Fix := Res and Fixed
End;

Begin
  If (ParamCount <> 1) or (ParamStr(1)='/?') then Begin
    WriteLn ('Usage: FIX_CRT.EXE <EXE_FILE> <'#13#10);
    WriteLn ('The program fixes sturtup Run-Time Error #200 caused by the CRT.Delay() bug.');
    WriteLn ('Backup file is created automatically.');
    WriteLn (#13#10'IMPORTANT NOTE: CRT.DELAY() ROUTINE WON''T BE WAITING DESIRED MILLISECONDS'#13#10+
                   '                AFTER PATCHING!!!'#13#10);
    WriteLn ('Written by Alexei A. Frounze (c) 1999. E-mail: alexfru@chat.ru');
    Exit
  End;
  Name := UpStr(ParamStr(1));
  If not Exists (Name) then Begin
    WriteLn ('File does not exist or couldn''t be opened successfully. Exiting...');
    Exit
  End;
  Prog := UpStr(ParamStr(0));
  FSplit (Name, D, N, E);
  If Pos(N+E, Prog) <> 0 then Begin
    WriteLn ('It''s not allowed to bugfix this program. Exiting...');
    Exit
  End;
  If CreateBackupCopy then WriteLn ('Backup copy has been created.')
  Else Begin
    WriteLn ('Backup copy has not been created successfully. Exiting...');
    Exit
  End;
  WriteLn ('Looking for the pattern to be changed. Please wait...');
  If Fix then WriteLn ('The bug has been fixed in the '+Name+' file.')
  Else WriteLn ('The bug has not been found or I/O error has occured.')
End.
